"""
Mesh module that will define the mesh.
"""


class Mesh(object):
    """define a mesh that will handle degrees-of-freedom (dof), element lengths
    etc."""

    def __init__(self, length, loads, reactions, dof):
        self._nodes = self.__get_nodes(length, loads, reactions)
        self._lengths = self.__get_lengths()
        self._num_elements = len(self.lengths)
        self._dof = dof * self.num_elements + dof

    @property
    def nodes(self):
        return self._nodes

    @property
    def dof(self):
        return self._dof

    @property
    def lengths(self):
        return self._lengths

    @property
    def num_elements(self):
        return self._num_elements

    def __get_lengths(self):
        # Calculate the lengths of each element
        lengths = []
        for k in range(len(self.nodes) - 1):
            lengths.append(self.nodes[k + 1] - self.nodes[k])
        return lengths

    def __get_nodes(self, length, loads, reactions):
        nodes = [0]  # ensure first node is always at zero (0)
        for param in loads + reactions:
            nodes.append(param.location)
        nodes.append(length)  # ensure last node is at the end of the beam
        nodes = list(set(nodes))   # remove duplicates
        nodes.sort()
        return nodes

    def __str__(self):
        s = ('MESH PARAMETERS\n'
            f'Number of elements: {self.num_elements}\n'
            f'Node locations: {self.nodes}\n'
            f'Element Lengths: {self.lengths}\n'
            f'Total degrees of freedom: {self.dof}\n')
        return s



"""
Module to define different reaction types
"""

from ._common import Forces


class Reaction(Forces):
    """Base class for all reactions"""

    name = None

    def __init__(self, location):
        super().__init__(value=None, location=location)
        self._force = None
        self._moment = None

    @property
    def force(self):
        return self._force

    @force.setter
    def force(self, f):
        self._force = f

    @property
    def moment(self):
        return self._moment

    @moment.setter
    def moment(self, m):
        self._moment = m

    @property
    def value(self):
        return (self.force, self.moment)

    def invalidate(self):
        """Invalidate the reaction values

        To be used whenever the parameters change and the reaction values are
        no longer valid
        """
        self.force, self.moment = (None, None)

    def __str__(self):
        return (f'{self.__class__.__name__}\n'
                f'  Location: {self.location}\n'
                f'     Force: {self.force}\n'
                f'    Moment: {self.moment}\n')

    def __repr__(self):
        return f'{self.__class__.__name__}(location={self.location})'


class PinnedReaction(Reaction):

    name = 'pinned'

    def __init__(self, location):
        super().__init__(location)
        # limit the vertical displacemnt but allow rotation
        self.boundary = (0, None)


class FixedReaction(Reaction):

    name = 'fixed'

    def __init__(self, location):
        super().__init__(location)
        # do not allow vertical or rotational displacement
        self.boundary = (0, 0)



"""module with exact equations for different load cases

https://www.awc.org/pdf/codes-standards/publications/design-aids/AWC-DA6-BeamFormulas-0710.pdf

"""

import numpy as np
import matplotlib.pyplot as plt

from ._base_elements import Base

from .reactions import PinnedReaction, FixedReaction
from .loads import PointLoad, MomentLoad


class ExactBase(Base):

    def __init__(self, L, E=1, Ixx=1):
        super().__init__(L, E, Ixx)
        self.EI = E * Ixx

    def deflection(self, x):
        raise NotImplemented('method must be overloaded!')
        # pass

    def plot(self, n=25):
        x = np.linspace(0, self.length, n)
        d = [self.deflection(xi) for xi in x]

        plt.plot(x, d)
        plt.show()


class SimpleSupportPointLoad(ExactBase):
    """Beam, simply supported at both ends, with a point load at any point
    between them.

    Case 8 from reference
    """

    def __init__(self, L, load, E=1, Ixx=1):
        super().__init__(L, E, Ixx)
        self.load = load
        self.reactions = [PinnedReaction(0), PinnedReaction(L)]

    def deflection(self, x):
        P = self.load.value
        L = self.length
        a = self.load.location
        b = L - a
        EI = self.EI
        if x < a:
            return P * b * x / (6 * EI * L) * (L**2 - b**2 - x**2)
        elif x > a:
            return P * a * (L - x) / (6 * EI * L) * (2 * L * x - x**2 - a**2)
        else:
            return P * a**2 * b**2 / (3 * EI * L)


def main():
    load = PointLoad(location=7, value=-1)
    b = SimpleSupportPointLoad(10, load)
    b.plot(700)


if __name__ == '__main__':
    main()



"""
Define a beam object that will have Loads and Reactions.
"""
from scipy.misc import derivative
import matplotlib.pyplot as plt
import numpy as np

# local imports
from ._base_elements import BeamElement
# from .reactions import PinnedReaction, FixedReaction


class Beam(BeamElement):
    """definition of a beam object"""

    def __init__(self, length, loads, reactions, E=1, Ixx=1):
        super().__init__(length, loads, reactions, E=E, Ixx=Ixx)

    def deflection(self, x):
        """calculate the deflection at the x location. Note that the x value
        given is the global x-value along the length of the beam.
        """

        # TODO: store the lengths/node locations in the class so they only have
        # to be assessed without recalculating
        nodes = self.mesh.nodes
        L, d = None, None

        # Using the given global x-value, determine the local x-value, length
        # of active element, and the nodal displacements (vertical, angular)
        # vector d
        for i in range(len(self.mesh.lengths)):
            if nodes[i] <= x and x <= nodes[i + 1]:
                # this is the element where the global x-value falls into.
                # Get the parameters in the local system and exit the loop
                x_local = x - nodes[i]
                # L = lengths[i]
                L = self.mesh.lengths[i]
                d = self.node_deflections[i * 2:i * 2 + 4]
                break

        if L is None or d is None:
            return
        return self.shape(x_local, L).dot(d)[0]

    def moment(self, x, dx=1):
        """calculate the moment in the beam at the global x value by taking
        the second derivative of the deflection curve.

        M(x) = E * Ixx * d^2 v(x) / dx^2
        """
        return self.E * self.Ixx * derivative(self.deflection, x, dx=dx, n=2)

    def shear(self, x, dx=1):
        """calculate the shear force at a given x location as the third
        derivative of displacement with respect to x

        V(x) = E * Ixx * d^3 v(x) / dx^3
        """
        return self.E * self.Ixx * derivative(self.deflection, x, dx=dx, n=3,
                                              order=5)

    def bending_stress(self, x, dx=1, c=1):
        """returns the bending stress at global coordinate x"""
        return self.moment(x, dx=dx) * c / self.Ixx

    def plot(self, n=250, plot_stress=True, title='Beam Analysis'):
        """plot the deflection, moment, and shear along the length of the beam
        """
        rows = 4 if plot_stress else 3
        fig, axes = plt.subplots(rows, 1, sharex=True)

        # locations of nodes in global coordinate system
        locations = self.mesh.nodes

        # Get the global x values. Note that the x-values for the moment and
        # shear do not contain the endpoints of the x values for the deflection
        # curve. This is because differentiation technique used is the central
        # difference formula, which cannot calculate the value at the
        # endpoints
        xd = np.linspace(0, self.length, n)  # deflection
        xm = xd[1:-2]                        # moment (and stress)
        xv = xm[2:-3]                        # shear
        v = [self.deflection(xi) for xi in xd]                  # deflection
        m = [self.moment(xi, dx=self.length / n) for xi in xm]  # moment
        V = [self.shear(xi, dx=self.length / n) for xi in xv]   # shear

        # Set up plotting variables to be able to iterate over them more easily
        xs = [xv, xm, xd]
        y = [V, m, v]
        labels = ['shear', 'moment', 'deflection']
        if plot_stress:
            q = [self.bending_stress(xi, dx=self.length / n) for xi in xm]
            xs.append(xm)
            y.append(q)
            labels.append('stress')

        for ax, x, y, label in zip(axes, xs, y, labels):
            ax.plot(x, y)
            ax.fill_between(x, y, 0, color='b', alpha=0.25)
            ax.set_ylabel(label)
            ax.grid(True)

        axes[-1].set_xlabel('Beam position, x')
        axes[-1].set_xticks(locations)

        fig.subplots_adjust(hspace=.25)
        fig.suptitle(title)
        return fig

    def __str__(self):
        L = ''
        for load in self.loads:
            L += 'Type: {}\n'.format(load.name)
            L += '    Location: {}\n'.format(load.location)
            L += '       Value: {}\n'.format(load.value)

        r = ''
        for reaction in self.reactions:
            r += 'Type: {}\n'.format(reaction.name)
            r += '    Location: {}\n'.format(reaction.location)
            r += '       Force: {}\n'.format(reaction.force)
            r += '      Moment: {}\n'.format(reaction.moment)

        msg = ('PARAMETERS\n'
               f'Length (length): {self.length}\n'
               f"Young's Modulus (E): {self.E}\n"
               f'Area moment of inertia (Ixx): {self.Ixx}\n'
                'LOADING\n'
               f'{L}\n'
               f'REACTIONS\n'
               f'{r}\n')
        return msg



"""
Base module that contains base classes to be used by other modules
"""


class Validator(object):
    """Decorator class used to validate parameters"""
    @staticmethod
    def positive(param_name='parameter'):
        """Function decorator to handle validating input parameters to ensure
        parameters are positive values.

        The input, param_name, is the parameter name that will show up in the
        call-stack when an invalid parameter is entered.
        """
        def decorator(func):
            def wrapper(*args, **kwargs):
                if args[1] <= 0:
                    raise ValueError(param_name + ' must be positive!')
                func(*args, **kwargs)
            return wrapper
        return decorator

    @staticmethod
    def non_negative(param_name='parameter'):
        """Function decorator to handle validating input parameters to ensure
        parameters are non-negative (positive or zero values).

        The input, param_name, is the parameter name that will show up in the
        call-stack when an invalid parameter is entered.
        """
        def decorator(func):
            def wrapper(*args, **kwargs):
                if args[1] < 0:
                    raise ValueError(param_name + ' must be non-negative!')
                func(*args, **kwargs)
            return wrapper
        return decorator

    @staticmethod
    def islist(param_name='parameter'):
        """Function decorator to handle validating input parameters to ensure
        parameters are a list.

        The input, param_name, is the parameter name that will show up in the
        call-stack when an invalid parameter is entered.
        """
        def decorator(func):
            def wrapper(*args, **kwargs):
                if not isinstance(args[1], list):
                    raise TypeError(param_name + ' must be a list!')
                func(*args, **kwargs)
            return wrapper
        return decorator


class Forces(object):
    """Base class for all loads and reactions"""

    def __init__(self, value, location=0):
        self.location = location
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def location(self):
        return self._location

    @location.setter
    @Validator.non_negative('location')
    def location(self, location):
        self._location = location

    def moment(self):
        return self.value * self.location

    def __repr__(self):
        return (f'{self.__class__.__name__}(value={self.value}, ' +
                                          f'location={self.location})')

    def __add__(self, force2):
        return self.value + force2.value

    def __sub__(self, load2):
        return self.value - load2.value



"""
Module to define different loads
"""

from ._common import Forces


class PointLoad(Forces):
    """
    class specific to a point load
    """

    name = 'point load'

    def __init__(self, value, location):
        super().__init__(value, location)


class MomentLoad(Forces):
    """
    class specific to a moment load
    """

    name = 'moment load'

    def __init__(self, value, location):
        super().__init__(value, location)



"""
Module to define a general mesh element to be used for any FEM element, and
the base element class that all FEM elements will be derived from
"""

import numpy as np
import matplotlib.pyplot as plt

from .mesh import Mesh
from ._common import Validator

# Importing loads is only used for checking the type. Find a better way to do
# this without needing to import loads
from .loads import PointLoad, MomentLoad


class Base(object):
    """base object to be used as base for both FEM analysis"""

    def __init__(self, length, E=1, Ixx=1):
        self.length = length
        self.E = E      # Young's modulus
        self.Ixx = Ixx  # area moment of inertia

    @property
    def length(self):
        return self._length

    @length.setter
    @Validator.positive('length')
    def length(self, length):
        self._length = length

    @property
    def E(self):
        return self._E

    @E.setter
    @Validator.positive("Young's modulus")
    def E(self, E):
        self._E = E

    @property
    def Ixx(self):
        return self._Ixx

    @Ixx.setter
    @Validator.positive('Area moment of inertia')
    def Ixx(self, Ixx):
        self._Ixx = Ixx


class Element(Base):
    """General element that will be inherited from for specific elements"""

    def __init__(self, length, E=1, Ixx=1):
        super().__init__(length, E, Ixx)
        self._node_deflections = None
        self._K = None  # global stiffness matrix

    def remesh(self):
        """force a remesh calculation and invalidate any calculation results"""
        raise NotImplemented('method must be overloaded')

    def invalidate(self):
        """invalidate the element to force resolving"""
        self._node_deflections = None
        self._K = None

    @property
    def K(self):
        """global stiffness matrix"""
        if self._K is None:
            self._K = self.stiffness_global()
        return self._K

    def solve(self):
        """solve the system the FEM system to define the nodal displacments
        and reaction forces.
        """
        self._calc_node_deflections()
        self._get_reaction_values()

    def _calc_node_deflections(self):
        raise NotImplemented('must be overloaded!')

    def _get_reaction_values(self):
        raise NotImplemented('must be overloaded!')

    def stiffness(self):
        """return local stiffness matrix, k, as numpy array evaluated with beam
        element length L, where L defaults to the length of the beam
        """
        raise NotImplemented('Method must be overloaded!')

    def stiffness_global(self):
        # Initialize the global stiffness matrix, then iterate over the
        # elements, calculate a local stiffness matrix, and add it to the
        # global stiffness matrix.
        raise NotImplemented('Method must be overloaded!')

    def apply_boundary_conditions(self, k, bcs):
        """
        Given the stiffness matrix 'k', and the boundary conditions as a list
        of tuples, apply the boundary conditions to the stiffness matrix by
        setting the rows and columns that correspond to the boundary conditions
        to zeros, with ones on the diagonal.

        The boundary conditions (bcs) are in the form
        bcs = [(displacement1, rotation1), (displacement2, rotation2)]

        For the boundary condition, if the conditional evaulates to None, then
        movement is allowed, otherwise no displacement is allowed.

        The boundary condition coordinates must match the stiffness matrix.
        That is, if the stiffness matrix is a local matrix, the boundary
        conditions must also be local.

        returns the adjusted  stiffness matrix after the boundary
        conditions are applied
        """

        def apply(k, i):
            """sub function to apply the boundary condition at row/col i to
            stiffness matrix k

            return the stiffness matrix k with boundary conditions applied
            """
            k[i] = 0     # set entire row to zeros
            k[:, i] = 0  # set entire column to zeros
            k[i][i] = 1  # set diagonal to 1
            return k

        # TODO: Check the sizes of the boundary conditions and stiffness matrix

        node = 0
        for bc in bcs:
            v, q = bc
            if v is not None:
                k = apply(k, node * 2)
            if q is not None:
                k = apply(k, node * 2 + 1)
            node += 1
        return k


class BeamElement(Element):
    """base beam element"""

    def __init__(self, length, loads, reactions, E=1, Ixx=1):
        super().__init__(length, E, Ixx)
        self.loads = loads
        self.reactions = reactions
        self.mesh = Mesh(length, loads, reactions, 2)

    def remesh(self):
        self.mesh = Mesh(self.length, self.loads, self.reactions, 2)
        self.invalidate()

    @property
    def loads(self):
        return self._loads

    @loads.setter
    @Validator.islist('loads')
    def loads(self, loads):
        self._loads = loads

    @property
    def reactions(self):
        return self._reactions

    @reactions.setter
    @Validator.islist('reactions')
    def reactions(self, reactions):
        self._reactions = reactions

    @property
    def node_deflections(self):
        if self._node_deflections is None:
            self._node_deflections = self._calc_node_deflections()
        return self._node_deflections

    def __get_boundary_conditions(self):
        # Initialize the  boundary conditions to None for each node, then
        # iterate over reactions and apply them to the boundary conditions
        # based on the reaction type.
        bc = [(None, None) for k in range(len(self.mesh.nodes))]
        for r in self.reactions:
            i = self.mesh.nodes.index(r.location)
            bc[i] = r.boundary
        return bc

    def _calc_node_deflections(self):
        """solve for vertical and angular displacement at each node"""

        # Get the boundary conditions from the reactions
        bc = self.__get_boundary_conditions()

        # Apply boundary conditions to global stiffness matrix. Note that the
        # boundary conditions are applied to a copy of the stiffness matrix to
        # avoid changing the property K, so it can still be used with further
        # calculations (ie, for calculating reaction values)
        kg = self.K.copy()
        kg = self.apply_boundary_conditions(kg, bc)

        # Use the same method of adding the input loads as the boundary
        # conditions. Start by initializing a numpy array to zero loads, then
        # iterate over the loads and add them to the appropriate index based on
        # the load type (force or moment)
        p = np.zeros((self.mesh.dof, 1))
        for ld in self.loads:
            i = self.mesh.nodes.index(ld.location)
            if isinstance(ld, PointLoad):
                p[i * 2][0] = ld.value  # input force
            elif isinstance(ld, MomentLoad):
                p[i * 2 + 1][0] = ld.value  # input moment

        # Solve the global system of equations {p} = [K]*{d} for {d}
        # save the deflection vector for the beam, so the analysis can be
        # reused without recalculating the stiffness matrix.
        # This vector should be cleared anytime any of the beam parameters
        # gets changed.
        self._node_deflections = np.linalg.solve(kg, p)
        return self._node_deflections

    def _get_reaction_values(self):
        """Calculate the nodal forces acting on the beam. Note that the forces
        will also include the input forces.

        reactions are calculated by solving the matrix equation
        {r} = [K] * {d}

        where
           - {r} is the vector of forces acting on the beam
           - [K] is the global stiffness matrix (without BCs applied)
           - {d} displacements of nodes
        """
        K = self.K                 # global stiffness matrix
        d = self.node_deflections  # force displacement vector
        r = np.matmul(K, d)

        for ri in self.reactions:
            i = self.mesh.nodes.index(ri.location)
            force, moment = r[i * 2: i * 2 + 2]

            # set the values in the reaction objects
            ri.force = force[0]
            ri.moment = moment[0]
        return r

    def shape(self, x, L=None):
        """return an array of the shape functions evaluated at x the local
        x-value
        """
        if L is None:
            L = self.length
        N1 = 1 / L**3 * (L**3 - 3 * L * x**2 + 2 * x**3)
        N2 = 1 / L**2 * (L**2 * x - 2 * L * x**2 + x**3)
        N3 = 1 / L**3 * (3 * L * x**2 - 2 * x**3)
        N4 = 1 / L**2 * (-L * x**2 + x**3)
        return np.array([N1, N2, N3, N4])

    def plot_shapes(self, n=25):
        """plot shape functions for the with n data points"""
        x = np.linspace(0, self.length, n)

        # set up list of axes with a grid where the two figures in each column
        # share an x axis
        axes = []
        fig = plt.figure()
        axes.append(fig.add_subplot(221))
        axes.append(fig.add_subplot(222))
        axes.append(fig.add_subplot(223, sharex=axes[0]))
        axes.append(fig.add_subplot(224, sharex=axes[1]))

        N = [[], [], [], []]
        for xi in x:
            n = self.shape(xi)
            for i in range(4):
                N[i].append(n[i])

        for k in range(4):
            ax = axes[k]
            ax.grid(True)
            ax.plot(x, N[k], label=f'$N_{k+1}$')
            ax.legend()

        fig.subplots_adjust(wspace=0.25, hspace=0)
        plt.show()

    def stiffness(self, L=None):
        """return local stiffness matrix, k, as numpy array evaluated with beam
        element length L, where L defaults to the length of the beam"""

        if L is None:
            L = self.length

        E = self.E
        Ixx = self.Ixx

        k = np.array([[12,       6 * L,       -12,        6 * L   ],
                      [6 * L,    4 * L**2,     -6 * L,    2 * L**2],
                      [-12,     -6 * L,        12,       -6 * L   ],
                      [6 * L,    2 * L**2,     -6 * L,    4 * L**2]])
        return E * Ixx / L**3 * k

    def stiffness_global(self):
        # Initialize the global stiffness matrix, then iterate over the
        # elements, calculate a local stiffness matrix, and add it to the
        # global stiffness matrix.
        kg = np.zeros((self.mesh.dof, self.mesh.dof))
        for e in range(self.mesh.num_elements):
            # iterate over all the elements and add the local stiffness matrix
            # to the global stiffness matrix at the proper index
            k = self.stiffness(self.mesh.lengths[e])  # local stiffness matrix
            i1, i2 = (e * 2, e * 2 + 4)               # global slicing index
            kg[i1:i2, i1:i2] = kg[i1:i2, i1:i2] + k   # current element
        self._K = kg

        return self._K



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='femethods',
    version='0.1.3dev',
    author='Joseph Contreras',
    author_email='26684136+JosephJContreras@users.noreply.github.com',
    description='Implementation of Finite Element Analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/josephjcontreras/FEMethods.git',
    packages=find_packages(),
    install_requires=['numpy', 'matplotlib', 'scipy'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
)
